# Collection of citations for modules in bbhx package

# Copyright (C) 2020 Michael L. Katz
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
:code:`bbhx.utils.citations`:

This module is used to collect citations for all modules in the package. This
module is then imported to add citations to module classes using their :code:`citation`
attribute.
"""

katz_citations = """
@article{Katz:2020hku,
    author = "Katz, Michael L. and Marsat, Sylvain and Chua, Alvin J. K. and Babak, Stanislav and Larson, Shane L.",
    title = "{GPU-accelerated massive black hole binary parameter estimation with LISA}",
    eprint = "2005.01827",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.102.023033",
    journal = "Phys. Rev. D",
    volume = "102",
    number = "2",
    pages = "023033",
    year = "2020"
}

@article{Katz:2021uax,
    author = "Katz, Michael L.",
    title = "{A fully-automated end-to-end pipeline for massive black hole binary signal extraction from LISA data}",
    eprint = "2111.01064",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    month = "11",
    year = "2021"
}

@article{michael katz_2021,
    title={mikekatz04/BBHx: First official public release},
    DOI={10.5281/zenodo.5730688},
    abstractNote={No description provided.},
    publisher={Zenodo},
    author={Michael Katz},
    year={2021},
    month={Nov}
}
"""

marsat_1 = """
@article{Marsat:2018oam,
    author = "Marsat, Sylvain and Baker, John G.",
    title = "{Fourier-domain modulations and delays of gravitational-wave signals}",
    eprint = "1806.10734",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    month = "6",
    year = "2018"
}
"""

marsat_2 = """
@article{Marsat:2020rtl,
    author = "Marsat, Sylvain and Baker, John G. and Dal Canton, Tito",
    title = "{Exploring the Bayesian parameter estimation of binary black holes with LISA}",
    eprint = "2003.00357",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.103.083011",
    journal = "Phys. Rev. D",
    volume = "103",
    number = "8",
    pages = "083011",
    year = "2021"
}
"""

phenomhm_citation = """
@article{London2018,
    archivePrefix = {arXiv},
    arxivId = {gr-qc/1708.00404},
    author = {London, Lionel and Khan, Sebastian and Fauchon-Jones, Edward and Garc\'ia, Cecilio and Hannam, Mark and Husa, Sascha and Jim{\'{e}}nez-Forteza, Xisco and Kalaghatgi, Chinmay and Ohme, Frank and Pannarale, Francesco},
    doi = {10.1103/PhysRevLett.120.161102},
    eprint = {1708.00404},
    journal = {\prl},
    keywords = {General Relativity and Quantum Cosmology},
    month = {apr},
    number = {16},
    pages = {161102},
    primaryClass = {gr-qc},
    title = {{First Higher-Multipole Model of Gravitational Waves from Spinning and Coalescing Black-Hole Binaries}},
    volume = {120},
    year = {2018}
}
"""

phenomd_citations = """
@article{Husa2016,
    archivePrefix = {arXiv},
    arxivId = {gr-qc/1508.07250},
    author = {Husa, S and Khan, S and Hannam, M and P{\"{u}}rrer, M and Ohme, F and Forteza, X.$\sim$J. and Boh{\'{e}}, A},
    doi = {10.1103/PhysRevD.93.044006},
    eprint = {1508.07250},
    journal = {\prd},
    month = {feb},
    number = {4},
    pages = {44006},
    primaryClass = {gr-qc},
    title = {{Frequency-domain gravitational waves from nonprecessing black-hole binaries. I. New numerical waveforms and anatomy of the signal}},
    volume = {93},
    year = {2016}
}


@article{Khan2016,
    archivePrefix = {arXiv},
    arxivId = {gr-qc/1508.07253},
    author = {Khan, S and Husa, S and Hannam, M and Ohme, F and P{\"{u}}rrer, M and Forteza, X.$\sim$J. and Boh{\'{e}}, A},
    doi = {10.1103/PhysRevD.93.044007},
    eprint = {1508.07253},
    journal = {\prd},
    month = {feb},
    number = {4},
    pages = {44007},
    primaryClass = {gr-qc},
    title = {{Frequency-domain gravitational waves from nonprecessing black-hole binaries. II. A phenomenological model for the advanced detector era}},
    volume = {93},
    year = {2016}
}
"""

Cornish_Heterodyning = """
@article{Cornish:2021lje,
    author = "Cornish, Neil J.",
    title = "{Heterodyned likelihood for rapid gravitational wave parameter inference}",
    eprint = "2109.02728",
    archivePrefix = "arXiv",
    primaryClass = "gr-qc",
    doi = "10.1103/PhysRevD.104.104054",
    journal = "Phys. Rev. D",
    volume = "104",
    number = "10",
    pages = "104054",
    year = "2021"
}
"""

Rel_Bin_citation = """
@article{Zackay:2018qdy,
    author = "Zackay, Barak and Dai, Liang and Venumadhav, Tejaswi",
    title = "{Relative Binning and Fast Likelihood Evaluation for Gravitational Wave Parameter Estimation}",
    eprint = "1806.08792",
    archivePrefix = "arXiv",
    primaryClass = "astro-ph.IM",
    month = "6",
    year = "2018"
}
"""
